{*******************************************************}
{                                                       }
{       Advanced SQL statement parser                   }
{       Copyright (c) 2001 AS Gaiasoft                  }
{       Created by Gert Kello                           }
{                                                       }
{*******************************************************}

unit gaSQLParserHelperClasses;

interface

uses
  gaAdvancedSQLParser, gaBasicSQLParser;

type
  TgaSQLStatementPart = class (TgaSQLTokenList)
  private
    FIsInvalid: Boolean;
    FParseComplete: Boolean;
    function GetParseComplete: Boolean;
    procedure SetIsInvalid(const Value: Boolean);
    procedure SetParseComplete(const Value: Boolean);
  protected
    procedure CheckModifyAllowed; virtual;
    function GetAsString: string; override;
    function InternalGetParseComplete: Boolean; virtual;
    procedure InternalSetParseComplete; virtual;
    procedure SetAsString(const Value: string); virtual;
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement); override;
    property AsString read GetAsString write SetAsString;
    property IsInvalid: Boolean read FIsInvalid write SetIsInvalid;
    property ParseComplete: Boolean read GetParseComplete write 
            SetParseComplete;
  end;
  
  TgaSQLExpression = class (TgaSQLStatementPart)
  private
    FIsCorrect: Boolean;
    FIsExpression: Boolean;
    FOldOnStatementAdd: TgaTokenEvent;
    FParenCount: Integer;
    function GetCanParseEnd: Boolean;
    function GetExpression: string;
  protected
    procedure AddParen(AToken: TgaSQLTokenObj); virtual;
    procedure DoAfterTokenAdd(AToken: TgaSQLTokenObj); reintroduce; override;
    procedure ParseExpression(AToken: TgaSQLTokenObj); virtual;
    procedure RemoveParen(AToken: TgaSQLTokenObj); virtual;
    procedure SetIsExpression(const Value: Boolean);
    property CanParseEnd: Boolean read GetCanParseEnd;
    property ParenCount: Integer read FParenCount write FParenCount;
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement); override;
    property Expression: string read GetExpression;
    property IsCorrect: Boolean read FIsCorrect;
    property IsExpression: Boolean read FIsExpression;
  end;
  
  TgaSQLStatementPartClass = class of TgaSQLStatementPart;

  TFieldExprParseState = (fepsNoSymbolsParsed, fepsFieldExpression,
    fepsFieldAlias, fepsExpression, fepsFieldAliasParsed, fepsFieldParseFinished);
  TgaSQLField = class (TgaSQLExpression)
  private
    FFieldAlias: TgaSQLTokenObj;
    FFieldName: TgaSQLTokenObj;
    FFieldParseState: TFieldExprParseState;
    FFieldPrefixies: TgaSQLTokenList;
    function GetFieldAlias: string;
    function GetFieldName: string;
    function GetFieldPrefix: string;
    procedure SetFieldAlias(const Value: string);
    procedure SetFieldName(const Value: string);
    procedure SetFieldPrefix(const Value: string);
  protected
    procedure DoAfterTokenAdd(AToken: TgaSQLTokenObj); override;
    procedure InternalSetParseComplete; override;
    procedure ParseFieldAlias(AToken: TgaSQLTokenObj);
    procedure ParseFieldExpression(AToken: TgaSQLTokenObj);
    procedure ParsePreFieldExpression(AToken: TgaSQLTokenObj);
    procedure StartExpressionParse;
    property FieldParseState: TFieldExprParseState read FFieldParseState;
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement); override;
    destructor Destroy; override;
    property FieldAlias: string read GetFieldAlias write SetFieldAlias;
    property FieldAliasToken: TgaSQLTokenObj read FFieldAlias;
    property FieldName: string read GetFieldName write SetFieldName;
    property FieldNameToken: TgaSQLTokenObj read FFieldName;
    property FieldPrefix: string read GetFieldPrefix write SetFieldPrefix;
    property FieldPrefixies: TgaSQLTokenList read FFieldPrefixies;
  end;
  
  TTableExprParseState = (tepsNoSymbolsParsed, tepsTableName, tepsTableAlias,
    tepsTableAliasParsed, tepsTableParseFinished);
  TgaSQLTable = class (TgaSQLStatementPart)
  private
    FIsAliasAllowed: Boolean;
    FTableAlias: TgaSQLTokenObj;
    FTableName: TgaSQLTokenObj;
    FTableParseState: TTableExprParseState;
    FTablePrefixies: TgaSQLTokenList;
    function GetTableAlias: string;
    function GetTableName: string;
    function GetTablePrefix: string;
    procedure SetTableAlias(const Value: string);
    procedure SetTableName(const Value: string);
    procedure SetTablePrefix(const Value: string);
  protected
    procedure DoAfterTokenAdd(AToken: TgaSQLTokenObj); override;
    procedure InternalSetParseComplete; override;
    procedure ParsePreTableName(AToken: TgaSQLTokenObj);
    procedure ParseTableAlias(AToken: TgaSQLTokenObj);
    procedure ParseTableName(AToken: TgaSQLTokenObj);
    property TableParseState: TTableExprParseState read FTableParseState write 
            FTableParseState;
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement); override;
    destructor Destroy; override;
    property IsAliasAllowed: Boolean read FIsAliasAllowed write FIsAliasAllowed;
    property TableAlias: string read GetTableAlias write SetTableAlias;
    property TableAliasToken: TgaSQLTokenObj read FTableAlias;
    property TableName: string read GetTableName write SetTableName;
    property TableNameToken: TgaSQLTokenObj read FTableName;
    property TablePrefix: string read GetTablePrefix write SetTablePrefix;
    property TablePrefixies: TgaSQLTokenList read FTablePrefixies;
  end;
  
  TgaSQLWhereExpression = class (TgaSQLExpression)
  end;
  
  TgaSQLStatementPartList = class (TgaListOfSQLTokenLists)
  private
    FCurrentPart: TgaSQLStatementPart;
    FOwnerStatement: TgaCustomSQLStatement;
    FStatementPartType: TgaSQLStatementPartClass;
  protected
    function GetAsString: string; override;
    function GetCurrentPart: TgaSQLStatementPart; virtual;
    procedure SetAsString(const Value: string); virtual;
    property OwnerStatement: TgaCustomSQLStatement read FOwnerStatement;
    property StatementPartType: TgaSQLStatementPartClass read 
            FStatementPartType;
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement; 
            AStatementPartType: TgaSQLStatementPartClass);
    procedure AddToken(AToken: TgaSQLTokenObj); override;
    procedure Clear; override;
    property AsString read GetAsString write SetAsString;
    property CurrentPart: TgaSQLStatementPart read GetCurrentPart;
  end;
  
  TgaSQLFieldList = class (TgaSQLStatementPartList)
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement);
  end;
  
  TgaSQLTableList = class (TgaSQLStatementPartList)
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement);
  end;
  
  TgaSQLOrderByList = class (TgaSQLFieldList)
  end;
  
procedure ParseStringToTokens(const AString: string; ATokenList: TgaSQLTokenList);

implementation

uses
  SysUtils, TypInfo, gaSQLParserConsts;

procedure ParseStringToTokens(const AString: string; ATokenList: TgaSQLTokenList);
var
  lSQLParser: TgaBasicSQLParser;
  lTokenObj: TgaSQLTokenObj;
begin
  ATokenList.Clear;
  lSQLParser := TgaBasicSQLParser.Create(nil);
  try
    lSQLParser.SQLText.Text := AString;
    while lSQLParser.TokenType <> stEnd do
    begin
      lTokenObj := ATokenList.NewToken;
      lTokenObj.AssignTokenInfo(lSQLParser);
      lSQLParser.NextToken;
    end;
  finally
    lSQLParser.Free;
  end;
end;

{
********************************* TgaSQLTable **********************************
}
constructor TgaSQLTable.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  inherited;
  FTablePrefixies := TgaSQLTokenList.Create(AOwnerStatement);
  FIsAliasAllowed := True;
  OwnerStatement.AllTables.Add(Self);
end;

destructor TgaSQLTable.Destroy;
begin
  FTablePrefixies.Free;
  OwnerStatement.AllTables.Remove(Self);
  inherited;
end;

procedure TgaSQLTable.DoAfterTokenAdd(AToken: TgaSQLTokenObj);
begin
  inherited;
  if AToken.TokenType = stComment then
    Exit;
  case TableParseState of
    tepsNoSymbolsParsed:
      ParsePreTableName(AToken);
    tepsTableName:
      ParseTableName(AToken);
    tepsTableAlias:
      ParseTableAlias(AToken);
    tepsTableAliasParsed:
      if not (AToken.TokenType in [stDelimitier, stComma, stEnd]) then
        IsInvalid := True;
    tepsTableParseFinished:
      raise Exception.Create(SerrTableParseFinised);
    { fepsExpression:
      Shouldn't get here as IsExpression should be true then }
    else
      raise Exception.CreateFmt(SerrUnecpectedFieldParseState,
        [GetEnumName(TypeInfo(TTableExprParseState), Ord(TableParseState))]);
  end;
  if (AToken.TokenType = stComma) then
    ParseComplete := True;
end;

function TgaSQLTable.GetTableAlias: string;
begin
  Result := GetTokenObjAsString(FTableAlias);
end;

function TgaSQLTable.GetTableName: string;
begin
  Result := GetTokenObjAsString(FTableName);
end;

function TgaSQLTable.GetTablePrefix: string;
begin
  Result := TablePrefixies.AsString;
end;

procedure TgaSQLTable.InternalSetParseComplete;
begin
  inherited;
  TableParseState := tepsTableParseFinished;
end;

procedure TgaSQLTable.ParsePreTableName(AToken: TgaSQLTokenObj);
begin
  case AToken.TokenType of
    stSymbol, stQuotedSymbol, stPeriod:
    begin
      FTableName := AToken;
      FTableParseState := tepsTableName;
    end;
    stDelimitier:
      {skip leading delimitiers};
    else
      IsInvalid := True;
  end;
end;

procedure TgaSQLTable.ParseTableAlias(AToken: TgaSQLTokenObj);
begin
  case AToken.TokenType of
    stDelimitier:
      {no special processing};
    stSymbol, stQuotedSymbol, stString:
      if not AToken.TokenSymbolIs('AS') then
      begin
        FTableAlias := AToken;
        FTableParseState := tepsTableAliasParsed;
      end;
    stOther:
      if AToken.TokenString = ';' then
        FTableParseState := tepsTableAliasParsed
      else
        IsInvalid := True;
    else
      IsInvalid := True;
  end;
  
end;

procedure TgaSQLTable.ParseTableName(AToken: TgaSQLTokenObj);
begin
  case AToken.TokenType of
    stSymbol, stQuotedSymbol, stPeriod:
    begin
      if FTableName <> nil then
        TablePrefixies.AddToken(AToken);
      FTableName := AToken;
    end;
    stDelimitier:
    begin
      if FTableName.TokenType = stPeriod then
        FTableName := nil;
      if IsAliasAllowed then
        FTableParseState := tepsTableAlias
      else
        FTableParseState := tepsTableAliasParsed;
    end;
    stComma:
      if FTableName.TokenType = stPeriod then
        FTableName := nil;
    stOther:
      if AToken.TokenString = ';' then
        FTableParseState := tepsTableAliasParsed
      else
        IsInvalid := True;
    else
      IsInvalid := True;
  end;
end;

procedure TgaSQLTable.SetTableAlias(const Value: string);
var
  tmpTokenList: TgaSQLTokenList;
  tmpToken: TgaSQLTokenObj;
begin
  if TableAlias <> Value then
  begin
    CheckModifyAllowed;
    if (not IsAliasAllowed) and (Value <> EmptyStr) then
      raise Exception.Create(SerrTableAliasNotAllowed);
    tmpTokenList := TgaSQLTokenList.Create(nil);
    try
      ParseStringToTokens(Value, tmpTokenList);
      TrimTokenList(tmpTokenList, True);
      if tmpTokenList.Count > 1 then
        raise Exception.CreateFmt(SerrWrongTokenCountInArg, ['Table alias', 1, tmpTokenList.Count]);
      if Assigned(FTableAlias) then
        OwnerStatement.ReplaceToken(FTableAlias, tmpTokenList.Items[0])
      else begin
        if not Assigned(FTableName) then
          raise Exception.Create(SerrNoTableForAlias);
        OwnerStatement.InsertTokenAfter(tmpTokenList.Items[0], FTableName);
        tmpToken := TgaSQLTokenObj.Create;
        tmpToken.SetTokenInfo(' ', stDelimitier, False, #0);
        OwnerStatement.InsertTokenAfter(tmpToken, FTableName);
      end;
      FTableAlias := tmpTokenList.Items[0];
    finally
      tmpTokenList.Free;
    end;
  end;
end;

procedure TgaSQLTable.SetTableName(const Value: string);
var
  tmpTokenList: TgaSQLTokenList;
begin
  if TableName <> Value then
  begin
    CheckModifyAllowed;
    tmpTokenList := TgaSQLTokenList.Create(nil);
    try
      ParseStringToTokens(Value, tmpTokenList);
      TrimTokenList(tmpTokenList, True);
      if tmpTokenList.Count <> 1 then
        raise Exception.CreateFmt(SerrWrongTokenCountInArg, ['Table name', 1, tmpTokenList.Count]);
      OwnerStatement.ReplaceToken(FTableName, tmpTokenList.Items[0]);
      FTableName := tmpTokenList.Items[0];
    finally
      tmpTokenList.Free;
    end;
  end;
end;

procedure TgaSQLTable.SetTablePrefix(const Value: string);
var
  tmpStr: string;
  tmpTokenList: TgaSQLTokenList;
  i: Integer;
begin
  if Value[Length(Value)] = '.' then
    tmpStr := Value
  else
    tmpStr := Value + '.';
  if (TablePrefix <> tmpStr) then
  begin
    CheckModifyAllowed;
    tmpTokenList := TgaSQLTokenList.Create(nil);
    try
      ParseStringToTokens(tmpStr, tmpTokenList);
      TrimTokenList(tmpTokenList, True);
      for i := 0 to TablePrefixies.Count - 1 do
        OwnerStatement.RemoveToken(TablePrefixies.Items[i]);
      TablePrefixies.Clear;
      for i := 0 to tmpTokenList.Count - 1 do
      begin
        OwnerStatement.InsertToken(tmpTokenList.Items[i], FTableName);
        TablePrefixies.AddToken(tmpTokenList.Items[i]);
      end;
    finally
      tmpTokenList.Free;
    end;
  end;
end;

{
********************************* TgaSQLField **********************************
}
constructor TgaSQLField.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  inherited;
  FFieldPrefixies := TgaSQLTokenList.Create(AOwnerStatement);
  SetIsExpression(False);
  OwnerStatement.AllFields.Add(Self);
end;

destructor TgaSQLField.Destroy;
begin
  FFieldPrefixies.Free;
  if Assigned(OwnerStatement) then
    OwnerStatement.AllFields.Remove(Self);
  inherited;
end;

procedure TgaSQLField.DoAfterTokenAdd(AToken: TgaSQLTokenObj);
begin
  inherited;
  { Field is written in SQL as follows:
    [table alias][.]FieldName[whitespace][FieldAlias],}
  if AToken.TokenType = stComment then
    Exit;
  if not IsExpression then
  begin
    case FieldParseState of
      fepsNoSymbolsParsed:
        ParsePreFieldExpression(AToken);
      fepsFieldExpression:
        ParseFieldExpression(AToken);
      fepsFieldAlias:
        ParseFieldAlias(AToken);
      fepsFieldAliasParsed:
        if AToken.TokenType <> stDelimitier then
          StartExpressionParse;
      fepsFieldParseFinished:
        raise Exception.Create(SerrFieldParseFinised);
      { fepsExpression:
        Shouldn't get here as IsExpression should be true then }
      else
        raise Exception.CreateFmt(SerrUnecpectedFieldParseState,
          [GetEnumName(TypeInfo(TFieldExprParseState), Ord(FieldParseState))]);
    end;
    if (AToken.TokenType = stComma) then
      ParseComplete := True;
  end;
end;

function TgaSQLField.GetFieldAlias: string;
begin
  Result := GetTokenObjAsString(FFieldAlias);
end;

function TgaSQLField.GetFieldName: string;
begin
  Result := GetTokenObjAsString(FFieldName);
end;

function TgaSQLField.GetFieldPrefix: string;
begin
  Result := FFieldPrefixies.AsString;
end;

procedure TgaSQLField.InternalSetParseComplete;
begin
  inherited;
  FFieldParseState := fepsFieldParseFinished;
end;

procedure TgaSQLField.ParseFieldAlias(AToken: TgaSQLTokenObj);
begin
  case AToken.TokenType of
    stDelimitier:
      {no special processing};
    stSymbol, stQuotedSymbol, stString:
      if not AToken.TokenSymbolIs('AS') then
      begin
        FFieldAlias := AToken;
        FFieldParseState := fepsFieldAliasParsed;
      end;
    else
      StartExpressionParse;
  end;
end;

procedure TgaSQLField.ParseFieldExpression(AToken: TgaSQLTokenObj);
begin
  case AToken.TokenType of
    stSymbol, stQuotedSymbol, stPeriod:
    begin
      if FFieldName <> nil then
        FieldPrefixies.AddToken(FFieldName);
      FFieldName := AToken;
    end;
    stDelimitier:
    begin
      if FFieldName.TokenType = stPeriod then
        FFieldName := nil;
      FFieldParseState := fepsFieldAlias;
    end;
    stComma:
      if FFieldName.TokenType = stPeriod then
        FFieldName := nil;
    stOther:
      if AToken.TokenString = '*' then
      begin
        if FFieldName <> nil then
          FieldPrefixies.AddToken(FFieldName);
        FFieldName := AToken;
      end else
        StartExpressionParse;
    else
      StartExpressionParse;
  end;
end;

procedure TgaSQLField.ParsePreFieldExpression(AToken: TgaSQLTokenObj);
begin
  case AToken.TokenType of
    stSymbol, stQuotedSymbol, stPeriod:
    begin
      FFieldName := AToken;
      FFieldParseState := fepsFieldExpression;
    end;
    stDelimitier:
      { skip leading whitespaces };
    stComma:
      IsInvalid := True;
    stOther:
      if AToken.TokenString = '*' then
      begin
        FFieldName := AToken;
        FFieldParseState := fepsFieldExpression;
      end else
        StartExpressionParse;
    else
      StartExpressionParse;
  end;
end;

procedure TgaSQLField.SetFieldAlias(const Value: string);
var
  tmpTokenList: TgaSQLTokenList;
  tmpToken, InsertAfterPoint: TgaSQLTokenObj;
begin
  if FieldAlias <> Value then
  begin
    CheckModifyAllowed;
    tmpTokenList := TgaSQLTokenList.Create(nil);
    try
      ParseStringToTokens(Value, tmpTokenList);
      TrimTokenList(tmpTokenList, True);
      if tmpTokenList.Count > 1 then
        raise Exception.CreateFmt(SerrWrongTokenCountInArg, ['Field alias', 1, tmpTokenList.Count]);
      if Assigned(FFieldAlias) then
        OwnerStatement.ReplaceToken(FFieldAlias, tmpTokenList.Items[0])
      else begin
        if IsExpression then
          InsertAfterPoint := Items[Count - 1]
        else
          InsertAfterPoint := FFieldName;
        if not Assigned(InsertAfterPoint) then
          raise Exception.Create(SerrNoInsPointForAlias);
        OwnerStatement.InsertTokenAfter(tmpTokenList.Items[0], InsertAfterPoint);
        tmpToken := TgaSQLTokenObj.Create;
        tmpToken.SetTokenInfo(' ', stDelimitier, False, #0);
        OwnerStatement.InsertTokenAfter(tmpToken, InsertAfterPoint);
      end;
      FFieldAlias := tmpTokenList.Items[0];
    finally
      tmpTokenList.Free;
    end;
  end;
end;

procedure TgaSQLField.SetFieldName(const Value: string);
var
  tmpTokenList: TgaSQLTokenList;
begin
  if FieldName <> Value then
  begin
    CheckModifyAllowed;
    if IsExpression then
      raise Exception.Create(SerrFieldAttrCantBeChangedInExpression);
    tmpTokenList := TgaSQLTokenList.Create(nil);
    try
      ParseStringToTokens(Value, tmpTokenList);
      TrimTokenList(tmpTokenList, True);
      if tmpTokenList.Count <> 1 then
        raise Exception.CreateFmt(SerrWrongTokenCountInArg, ['Field name', 1, tmpTokenList.Count]);
      OwnerStatement.ReplaceToken(FFieldName, tmpTokenList.Items[0]);
      FFieldName := tmpTokenList.Items[0];
    finally
      tmpTokenList.Free;
    end;
  end;
end;

procedure TgaSQLField.SetFieldPrefix(const Value: string);
var
  tmpStr: string;
  tmpTokenList: TgaSQLTokenList;
  i: Integer;
begin
  if Value[Length(Value)] = '.' then
    tmpStr := Value
  else
    tmpStr := Value + '.';
  if (FieldPrefix <> tmpStr) then
  begin
    CheckModifyAllowed;
    if IsExpression then
      raise Exception.Create(SerrFieldAttrCantBeChangedInExpression);
    tmpTokenList := TgaSQLTokenList.Create(nil);
    try
      ParseStringToTokens(tmpStr, tmpTokenList);
      TrimTokenList(tmpTokenList, True);
      for i := 0 to FieldPrefixies.Count - 1 do
        OwnerStatement.RemoveToken(FieldPrefixies.Items[i]);
      FieldPrefixies.Clear;
      for i := 0 to tmpTokenList.Count - 1 do
      begin
        OwnerStatement.InsertToken(tmpTokenList.Items[i], FFieldName);
        FieldPrefixies.AddToken(tmpTokenList.Items[i]);
      end;
    finally
      tmpTokenList.Free;
    end;
  end;
end;

procedure TgaSQLField.StartExpressionParse;
var
  i: Integer;
begin
  FFieldName := nil;
  FFieldAlias := nil;
  FFieldPrefixies.Clear;
  FIsExpression := True;
  FFieldParseState := fepsExpression;
  for i := 0 to Count - 1 do
    ParseExpression(Items[i]);
end;

{
*************************** TgaSQLStatementPartList ****************************
}
constructor TgaSQLStatementPartList.Create(AOwnerStatement: 
        TgaCustomSQLStatement; AStatementPartType: TgaSQLStatementPartClass);
begin
  inherited Create;
  FStatementPartType := AStatementPartType;
  FOwnerStatement := AOwnerStatement;
end;

procedure TgaSQLStatementPartList.AddToken(AToken: TgaSQLTokenObj);
begin
  CurrentPart.AddToken(AToken);
  if CurrentPart.ParseComplete then
    FCurrentPart := nil;
end;

procedure TgaSQLStatementPartList.Clear;
begin
  FCurrentPart := nil;
  inherited Clear;
end;

function TgaSQLStatementPartList.GetAsString: string;
begin
  Result := inherited GetAsString;
end;

function TgaSQLStatementPartList.GetCurrentPart: TgaSQLStatementPart;
begin
  if not Assigned(FCurrentPart) then
  begin
    FCurrentPart := StatementPartType.Create(OwnerStatement);
    Add(FCurrentPart);
  end;
  Result := FCurrentPart;
end;

procedure TgaSQLStatementPartList.SetAsString(const Value: string);
var
  tmpTokenList: TgaSQLTokenList;
  tmpRemoveTokenList: TgaSQLTokenList;
  i: Integer;
begin
  tmpTokenList := TgaSQLTokenList.Create(nil);
  try
    tmpTokenList := TgaSQLTokenList.Create(nil);
    ParseStringToTokens(Value, tmpTokenList);
    TrimTokenList(tmpTokenList, True);
    tmpRemoveTokenList := TgaSQLTokenList.Create(nil);
    try
      GetAllTokens(tmpRemoveTokenList);
      TrimTokenList(tmpRemoveTokenList, False);
      OwnerStatement.ReplaceTokenList(tmpTokenList, tmpRemoveTokenList);
    finally
      tmpRemoveTokenList.Free;
    end;
    Clear;
    for i := 0 to tmpTokenList.Count - 1 do
      AddToken(tmpTokenList[i]);
  finally
    tmpTokenList.Free;
  end;
end;

{
******************************* TgaSQLFieldList ********************************
}
constructor TgaSQLFieldList.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  inherited Create(AOwnerSTatement, TgaSQLField);
end;

{
******************************* TgaSQLTableList ********************************
}
constructor TgaSQLTableList.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  inherited Create(AOwnerSTatement, TgaSQLTable);
end;

{
***************************** TgaSQLStatementPart ******************************
}
constructor TgaSQLStatementPart.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  if not Assigned(AOwnerStatement) then
    raise Exception.Create(SerrStmPartWithoutStm);
  inherited;
end;

procedure TgaSQLStatementPart.CheckModifyAllowed;
begin
  if not Assigned(OwnerStatement) then
    raise Exception.Create(SerrNowOwnerParser)
end;

function TgaSQLStatementPart.GetAsString: string;
begin
  Result := inherited GetAsString;
end;

function TgaSQLStatementPart.GetParseComplete: Boolean;
begin
  Result := InternalGetParseComplete;
end;

function TgaSQLStatementPart.InternalGetParseComplete: Boolean;
begin
  Result := FParseComplete;
end;

procedure TgaSQLStatementPart.InternalSetParseComplete;
begin
  FParseComplete := True;
end;

procedure TgaSQLStatementPart.SetAsString(const Value: string);
var
  tmpTokenList: TgaSQLTokenList;
  i: Integer;
  tmpRemoveTokenList: TgaSQLTokenList;
begin
  CheckModifyAllowed;
  tmpTokenList := TgaSQLTokenList.Create(nil);
  try
    ParseStringToTokens(Value, tmpTokenList);
    TrimTokenList(tmpTokenList, True);
    tmpRemoveTokenList := TgaSQLTokenList.Create(nil);
    try
      GetAllTokens(tmpRemoveTokenList);
      TrimTokenList(tmpRemoveTokenList, False);
      OwnerStatement.ReplaceTokenList(tmpTokenList, tmpRemoveTokenList);
    finally
      tmpRemoveTokenList.Free;
    end;
    Clear;
    for i := 0 to tmpTokenList.Count - 1 do
      AddToken(tmpTokenList[i]);
  finally
    tmpTokenList.Free;
  end;
end;

procedure TgaSQLStatementPart.SetIsInvalid(const Value: Boolean);
begin
  FIsInvalid := Value;
end;

procedure TgaSQLStatementPart.SetParseComplete(const Value: Boolean);
begin
  if Value then
    InternalSetParseComplete;
end;

{
******************************* TgaSQLExpression *******************************
}
constructor TgaSQLExpression.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  inherited Create(AOwnerStatement);
  FisExpression := True;
  FIsCorrect := True;
end;

procedure TgaSQLExpression.AddParen(AToken: TgaSQLTokenObj);
begin
  if FParenCount = 0 then
  begin
    FOldOnStatementAdd := OwnerStatement.OnTokenAdded;
    OwnerStatement.OnTokenAdded := DoAddToken;
  end;
  Inc(FParenCount);
end;

procedure TgaSQLExpression.DoAfterTokenAdd(AToken: TgaSQLTokenObj);
begin
  inherited ;
  if AToken.TokenType = stComment then
    Exit;
  if IsExpression then
    ParseExpression(AToken);
end;

function TgaSQLExpression.GetCanParseEnd: Boolean;
begin
  Result := ParenCount = 0;
end;

function TgaSQLExpression.GetExpression: string;
begin
  if IsExpression then
    Result := AsString
  else
    Result := EmptyStr;
end;

procedure TgaSQLExpression.ParseExpression(AToken: TgaSQLTokenObj);
begin
  // #Todo3 Expression parsing not implemented... yet
  case AToken.TokenType of
    stLParen:
      AddParen(AToken);
    stRParen:
      RemoveParen(AToken);
  end;
end;

procedure TgaSQLExpression.RemoveParen(AToken: TgaSQLTokenObj);
begin
  Dec(FParenCount);
  if ParenCount = 0 then
    OwnerStatement.OnTokenAdded := FOldOnStatementAdd;
  if ParenCount < 0 then
    FIsCorrect := False;
end;

procedure TgaSQLExpression.SetIsExpression(const Value: Boolean);
begin
  FIsExpression := Value;
end;

end.
